from __future__ import annotations

import httpx

from typing import Optional, Literal

__all__ = [
    "BadRequestError",
    "AuthenticationError",
    "PermissionDeniedError",
    "NotFoundError",
    "ConflictError",
    "UnprocessableEntityError",
    "RateLimitError",
    "InternalServerError",
]


class TwelveLabsError(Exception):
    pass


class APIError(TwelveLabsError):
    message: str
    request: httpx
    body: Optional[object]

    def __init__(
        self, message: str, request: httpx.Request, *, body: Optional[object]
    ) -> None:
        super().__init__(message)
        self.message = message
        self.request = request
        self.body = body


class APIStatusError(APIError):
    """Raised API response return a status code of 4xx or 5xx."""

    response: httpx.Response
    status_code: int

    def __init__(
        self, message: str, *, response: httpx.Response, body: Optional[object]
    ) -> None:
        super().__init__(message, response.request, body=body)
        self.response = response
        self.status_code = response.status_code


class APIConnectionError(APIError):
    def __init__(
        self, *, message: str = "Failed connection", request: httpx.Request
    ) -> None:
        super().__init__(message, request, body=None)


class APITimeoutError(APIConnectionError):
    def __init__(self, *, request: httpx.Request) -> None:
        super().__init__(message="Request timed out", request=request)


class BadRequestError(APIStatusError):
    status_code: Literal[400] = 400


class AuthenticationError(APIStatusError):
    status_code: Literal[401] = 401


class PermissionDeniedError(APIStatusError):
    status_code: Literal[403] = 403


class NotFoundError(APIStatusError):
    status_code: Literal[404] = 404


class ConflictError(APIStatusError):
    status_code: Literal[409] = 409


class UnprocessableEntityError(APIStatusError):
    status_code: Literal[422] = 422


class RateLimitError(APIStatusError):
    status_code: Literal[429] = 429


class InternalServerError(APIStatusError):
    pass



from ._base import Object, ObjectWithTimestamp, PageInfo, TokenPageInfo, RootModelList
from .index import Index, IndexListWithPagination
from .task import (
    Task,
    TaskStatus,
    TaskListWithPagination,
    TransferImportResponse,
    TransferImportStatusResponse,
    TransferImportLog,
)
from .video import Video, VideoValue, VideoListWithPagination
from .search import (
    SearchData,
    SearchResult,
    CombinedSearchResult,
)
from .generate import (
    GenerateOpenEndedTextResult,
    GenerateSummarizeChapterResult,
    GenerateSummarizeHighlightResult,
    GenerateSummarizeResult,
    GenerateOpenEndedTextStreamResult,
)
from .embed import (
    CreateEmbeddingsResult,
    EmbeddingsTask,
    EmbeddingsTaskStatus,
    CreateEmbeddingsTaskVideoParams,
    EmbeddingsTaskListWithPagination,
    Embedding,
    SegmentEmbedding,
)

__all__ = [
    "Object",
    "ObjectWithTimestamp",
    "PageInfo",
    "TokenPageInfo",
    "Index",
    "IndexListWithPagination",
    "Task",
    "TaskListWithPagination",
    "TaskStatus",
    "TransferImportResponse",
    "TransferImportStatusResponse",
    "TransferImportLog",
    "Video",
    "VideoValue",
    "VideoListWithPagination",
    "SearchData",
    "SearchResult",
    "CombinedSearchResult",
    "GenerateOpenEndedTextResult",
    "GenerateSummarizeChapterResult",
    "GenerateSummarizeHighlightResult",
    "GenerateSummarizeResult",
    "GenerateOpenEndedTextStreamResult",
    "CreateEmbeddingsResult",
    "EmbeddingsTask",
    "EmbeddingsTaskStatus",
    "CreateEmbeddingsTaskVideoParams",
    "EmbeddingsTaskListWithPagination",
    "Embedding",
    "SegmentEmbedding",
    "RootModelList",
]



import httpx

BASE_URL = "https://api.twelvelabs.io"
API_KEY_HEADER = "x-api-key"
LATEST_API_VERSION = "v1.3"

# default timeout is 10 minutes
DEFAULT_TIMEOUT = httpx.Timeout(timeout=10 * 60.0, connect=5.0)
DEFAULT_MAX_RETRIES = 2
DEFAULT_LIMITS = httpx.Limits(max_connections=100, max_keepalive_connections=20)



import httpx
import json

from . import exceptions
from .constants import DEFAULT_TIMEOUT, API_KEY_HEADER
from .exceptions import APIStatusError, APITimeoutError
from typing import Optional, Dict, Any


class APIClient:
    _client: httpx.Client

    def __init__(
        self,
        base_url: str,
        api_key: str,
        *,
        timeout: Optional[httpx.Timeout] = DEFAULT_TIMEOUT,
    ) -> None:
        httpClient = httpx.Client(
            base_url=base_url,
            headers={API_KEY_HEADER: api_key},
            timeout=timeout,
        )
        self._client = httpClient

    def _request(self, method: str, url: str, **kwargs) -> Any:
        response = None
        try:
            response = self._client.request(method, url, **kwargs)
            response.raise_for_status()
        except httpx.TimeoutException as e:
            raise APITimeoutError(request=getattr(response, "_request", None))
        except httpx.HTTPStatusError as e:
            raise self._make_status_error(e.response)
        except Exception as e:
            raise exceptions.APIConnectionError(
                request=getattr(response, "_request", None)
            )

        if len(response.content) > 0 and "application/json" in response.headers.get(
            "Content-Type", ""
        ):
            return response.json()
        return response.text

    def _request_stream(self, method: str, url: str, **kwargs):
        return self._client.stream(method, url, **kwargs)

    def _make_status_error(self, response: httpx.Response) -> APIStatusError:
        if response.is_closed and not response.is_stream_consumed:
            body = None
            err_msg = f"Error code: {response.status_code}"
        else:
            err_text = response.text.strip()
            body = err_text

            try:
                body = json.loads(err_text)
                err_msg = f"Error code: {response.status_code} - {body}"
            except Exception:
                err_msg = err_text or f"Error code: {response.status_code}"

        if response.status_code == 400:
            return exceptions.BadRequestError(err_msg, response=response, body=body)

        if response.status_code == 401:
            return exceptions.AuthenticationError(err_msg, response=response, body=body)

        if response.status_code == 403:
            return exceptions.PermissionDeniedError(
                err_msg, response=response, body=body
            )

        if response.status_code == 404:
            return exceptions.NotFoundError(err_msg, response=response, body=body)

        if response.status_code == 409:
            return exceptions.ConflictError(err_msg, response=response, body=body)

        if response.status_code == 422:
            return exceptions.UnprocessableEntityError(
                err_msg, response=response, body=body
            )

        if response.status_code == 429:
            return exceptions.RateLimitError(err_msg, response=response, body=body)

        if response.status_code >= 500:
            return exceptions.InternalServerError(err_msg, response=response, body=body)
        return APIStatusError(err_msg, response=response, body=body)

    def get(self, url: str, params: Optional[Dict[str, Any]] = None, **kwargs) -> Any:
        return self._request("GET", url, params=params, **kwargs)

    def post(
        self,
        url: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        *,
        stream: bool = False,
        **kwargs,
    ) -> Any:
        if stream:
            return self._request_stream("POST", url, data=data, json=json, **kwargs)
        return self._request("POST", url, data=data, json=json, **kwargs)

    def patch(
        self,
        url: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Any:
        return self._request("PATCH", url, data=data, json=json, **kwargs)

    def put(
        self,
        url: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Any:
        return self._request("PUT", url, data=data, json=json, **kwargs)

    def delete(self, url: str, **kwargs) -> Any:
        return self._request("DELETE", url, **kwargs)



from typing import List, Optional, Literal, Union, Dict, Any, BinaryIO
from io import BytesIO

from ..resource import APIResource
from .. import models
from ..util import remove_none_values

import json as jsonutil


class Search(APIResource):
    def query(
        self,
        index_id: str,
        options: List[Literal["visual", "audio"]],
        *,
        query_text: str = None,
        query_media_type: Literal["image"] = None,
        query_media_file: Union[str, BinaryIO, None] = None,
        query_media_url: str = None,
        group_by: Optional[Literal["video", "clip"]] = None,
        threshold: Optional[Literal["high", "medium", "low", "none"]] = None,
        operator: Optional[Literal["or", "and"]] = None,
        filter: Optional[Dict[str, Any]] = None,
        page_limit: Optional[int] = None,
        sort_option: Optional[Literal["score", "clip_count"]] = None,
        adjust_confidence_level: Optional[float] = None,
        **kwargs,
    ) -> models.SearchResult:
        if not query_text and not query_media_file and not query_media_url:
            raise ValueError("Either `query_text`, `query_media_file` or `query_media_url` must be provided.")
        
        if filter is not None:
            filter = jsonutil.dumps(filter)
        data = {
            "index_id": index_id,
            "query_text": query_text,
            "query_media_type": query_media_type,
            "query_media_url": query_media_url,
            "search_options": options,
            "group_by": group_by,
            "threshold": threshold,
            "operator": operator,
            "filter": filter,
            "page_limit": page_limit,
            "sort_option": sort_option,
            "adjust_confidence_level": adjust_confidence_level,
        }

        files = {}
        opened_files: List[BinaryIO] = []
        if query_media_file is not None:
            if isinstance(query_media_file, str):
                file = open(query_media_file, "rb")
                opened_files.append(file)
                files["query_media_file"] = file
            else:
                files["query_media_file"] = query_media_file
        if len(files) == 0:
            files["_"] = "" # dummy file to make the form data valid

        try:
            res = self._post(
                "search", data=remove_none_values(data), files=files, **kwargs
            )
            return models.SearchResult(self, **res)
        finally:
            for file in opened_files:
                file.close()

    def by_page_token(self, page_token: str, **kwargs) -> models.SearchResult:
        res = self._get(f"search/{page_token}", **kwargs)
        return models.SearchResult(self, **res)



from __future__ import annotations

from typing import Optional, Dict, Any, Callable, List, TYPE_CHECKING
from pydantic import PrivateAttr

from ._base import ObjectWithTimestamp, BaseModel, ModelMixin, PageInfo, RootModelList

if TYPE_CHECKING:
    from ..resources import Task as TaskResource


class TaskHLS(BaseModel):
    video_url: Optional[str] = None
    thumbnail_urls: Optional[List[str]] = None
    status: Optional[str] = None
    updated_at: Optional[str] = None


class Task(ObjectWithTimestamp):
    _resource: TaskResource = PrivateAttr()
    index_id: str
    video_id: Optional[str] = None
    status: str
    system_metadata: Dict[str, Any]
    hls: Optional[TaskHLS] = None

    def __init__(self, resource: TaskResource, **data):
        super().__init__(**data)
        self._resource = resource

    # Task related methods

    @property
    def done(self) -> bool:
        return self.status in ("ready", "failed")

    def retrieve(self, **kwargs) -> Task:
        return self._resource.retrieve(self.id, **kwargs)

    def delete(self, **kwargs) -> None:
        return self._resource.delete(self.id, **kwargs)

    def wait_for_done(
        self,
        *,
        sleep_interval: float = 5.0,
        callback: Optional[Callable[[Task], None]] = None,
        **kwargs,
    ) -> Task:
        if sleep_interval <= 0:
            raise ValueError("sleep_interval must be greater than 0")
        while not self.done:
            self._resource._sleep(sleep_interval)
            try:
                task = self.retrieve(**kwargs)
                self.status = task.status
                self.system_metadata = task.system_metadata
            except Exception as e:
                print(f"Retrieving task failed: {e}. Retrying..")
                continue
            if callback is not None:
                callback(self)
        return self


class TaskListWithPagination(ModelMixin, BaseModel):
    _resource: TaskResource = PrivateAttr()
    _origin_params: Dict[str, Any] = PrivateAttr()
    data: RootModelList[Task] = []
    page_info: PageInfo

    def __init__(self, resource: TaskResource, origin_params: Dict[str, Any], **data):
        super().__init__(**data)
        self._resource = resource
        self._origin_params = origin_params

    def __iter__(self):
        return self

    def __next__(self) -> RootModelList[Task]:
        if self.page_info.page >= self.page_info.total_page:
            raise StopIteration
        params = self._origin_params
        params["page"] = self.page_info.page + 1
        res = self._resource.list_pagination(**params)
        self.page_info = res.page_info
        return res.data


class TaskStatus(BaseModel):
    index_id: str
    ready: int
    validating: int
    pending: int
    failed: int
    total_result: int


# Transfer Import


class TransferImportVideo(BaseModel):
    video_id: str
    filename: str


class TransferImportFailedFile(BaseModel):
    filename: str
    error_message: str


class TransferImportResponse(BaseModel):
    videos: List[TransferImportVideo] = []
    failed_files: Optional[List[TransferImportFailedFile]] = None


# Transfer Import Status


class TransferImportStatusVideoInfo(BaseModel):
    video_id: str
    filename: str
    created_at: Optional[str] = None


class TransferImportStatusFailedVideoInfo(BaseModel):
    filename: str
    created_at: Optional[str] = None
    error_message: str


class TransferImportStatusResponse(BaseModel):
    not_imported: List[str] = []
    ready: List[TransferImportStatusVideoInfo] = []
    validating: List[TransferImportStatusVideoInfo] = []
    pending: List[TransferImportStatusVideoInfo] = []
    indexing: List[TransferImportStatusVideoInfo] = []
    failed: List[TransferImportStatusFailedVideoInfo] = []
    queued: List[TransferImportStatusVideoInfo] = []


# Transfer Import Log


class TransferImportLogVideoStatus(BaseModel):
    ready: int
    validating: int
    pending: int
    indexing: int
    failed: int
    queued: int


class TransferImportLogFailedFile(BaseModel):
    filename: str
    error_message: str


class TransferImportLog(BaseModel):
    index_id: str
    index_name: str
    created_at: str
    ended_at: Optional[str] = None
    video_status: TransferImportLogVideoStatus
    failed_files: Optional[List[TransferImportLogFailedFile]] = None



from setuptools import setup, find_packages
from pathlib import Path
import os

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
version_file = os.path.join(os.path.dirname(__file__), "VERSION")
with open(version_file, "r") as f:
    version = f.read().strip()

with open("requirements.txt") as f:
    required = f.read().splitlines()

setup(
    name="twelvelabs",
    version=version,
    author="Twelve Labs",
    description="SDK for Twelve Labs API",
    url="https://github.com/twelvelabs-io/twelvelabs-python",
    packages=find_packages(),
    install_requires=required,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
